"""Hive Device Attribute Module."""
from .custom_logging import Logger
from .hive_data import Data


class Attributes:
    """Device Attributes Code."""

    def __init__(self):
        self.log = Logger()
        self.type = "Attribute"

    async def state_attributes(self, n_id, _type):
        """Get HA State Attributes"""
        from .hive_session import Session
        await self.log.log(n_id, "Extra", "Getting Attribute data")
        attr = {}

        if n_id in Data.products or n_id in Data.devices:
            attr.update({"available": (await self.online_offline(n_id))})
            if n_id in Data.BATTERY:
                battery = await self.battery(n_id)
                if battery is not None:
                    attr.update({"battery": str(battery) + "%"})
            if n_id in Data.MODE:
                attr.update({"mode": (await self.get_mode(n_id))})

            if _type in Data.HIVE_TYPES["Sensor"]:
                data = Data.products[n_id]
                rec = data["props"].get("presenceLastChanged", False)
                if rec:
                    trim = "{:10.10}".format(str(rec))
                    time = await Session.epochtime(
                        trim, "%d-%m-%Y %H:%M:%S", "from_epoch")
                    attr.update({"state_changed": time})

        await self.log.log(n_id, "Extra", "Attribute data {0}", info=attr)
        return attr

    async def online_offline(self, n_id):
        """Check if device is online"""
        state = None
        final = None

        if n_id in Data.devices:
            data = Data.devices[n_id]
            state = data["props"]["online"]
            final = state
            Data.NODES[n_id]["Availabile"] = final
        else:
            await self.log.error_check(n_id, "ERROR", "Failed")

        return final if final is None else Data.NODES[n_id]["Availabile"]

    async def get_mode(self, n_id):
        """Get sensor mode."""
        online = await self.online_offline(n_id)
        state = None
        final = None

        if n_id in Data.products:
            if online:
                data = Data.products[n_id]
                state = data["state"]["mode"]
            await self.log.error_check(n_id, self.type, state)
            final = Data.HIVETOHA[self.type].get(state, state)
            Data.NODES[n_id]["Device_Mode"] = final
        else:
            await self.log.error_check(n_id, "ERROR", "Failed")

        return final if final is None else Data.NODES[n_id]["Device_Mode"]

    async def battery(self, n_id):
        """Get device battery level."""
        online = await self.online_offline(n_id)
        state = None
        final = None

        if n_id in Data.devices:
            if online:
                data = Data.devices[n_id]
                state = data["props"]["battery"]
                final = state
                Data.NODES[n_id]["BatteryLevel"] = final
            await self.log.error_check(n_id, self.type, state)
        else:
            await self.log.error_check(n_id, "ERROR", "Failed")

        return final if final is None else Data.NODES[n_id]["BatteryLevel"]



"""Hive Switch Module."""
import asyncio
from typing import Optional
from aiohttp import ClientSession

from .hive_session import Session
from .hive_data import Data
from .custom_logging import Logger
from .device_attributes import Attributes
from .hive_async_api import Hive_Async


class Plug():
    """Hive Switch Code."""

    def __init__(self, websession: Optional[ClientSession] = None):
        """Initialise."""
        self.hive = Hive_Async(websession)
        self.session = Session(websession)
        self.log = Logger()
        self.attr = Attributes()
        self.type = "Switch"

    async def get_plug(self, device):
        """Get smart plug data."""
        await self.log.log(device["hive_id"], self.type, "Getting switch data.")
        online = await self.attr.online_offline(device["hive_id"])
        error = await self.log.error_check(device["hive_id"], self.type, online)

        dev_data = {}
        if device["hive_id"] in Data.devices:
            data = Data.devices[device["hive_id"]]
            dev_data = {"hive_id": device["hive_id"],
                        "hive_name": device["hive_name"],
                        "hive_type": device["hive_type"],
                        "ha_name": device["ha_name"],
                        "ha_type": device["ha_type"],
                        "device_id": device["device_id"],
                        "device_name": device["device_name"],
                        "state": await self.get_state(device),
                        "power_usage": await self.get_power_usage(device),
                        "device_data": data.get("props", None),
                        "parent_device": data.get("parent", None),
                        "custom": device.get("custom", None),
                        "attributes": await self.attr.state_attributes(device["hive_id"],
                                                                       device["hive_type"])
                        }

        if not error:
            await self.log.log(device["hive_id"], self.type,
                               "Device update {0}", info=dev_data)

        return dev_data

    async def get_state(self, device):
        """Get plug current state."""
        await self.log.log(device["hive_id"], "Extra", "Getting state of switch")
        online = await self.attr.online_offline(device["hive_id"])
        state = None
        final = None

        if device["hive_id"] in Data.products:
            if online:
                data = Data.products[device["hive_id"]]
                state = data["state"]["status"]
                await self.log.log(device["hive_id"], "Extra", "Status is {0}", info=state)
            await self.log.error_check(device["hive_id"], "Extra", online)
            final = Data.HIVETOHA["Switch"].get(state, state)
            Data.NODES[device["hive_id"]]["State"] = final
        else:
            await self.log.error_check(device["hive_id"], "ERROR", "Failed")

        return final if final is None else Data.NODES[device["hive_id"]]["State"]

    async def get_power_usage(self, device):
        """Get smart plug current power usage."""
        await self.log.log(device["hive_id"], "Extra", "Getting power consumption.")
        online = await self.attr.online_offline(device["hive_id"])
        state = None
        final = None

        if device["hive_id"] in Data.products:
            if online:
                data = Data.products[device["hive_id"]]
                state = data["props"]["powerConsumption"]
                await self.log.log(device["hive_id"], "Extra",
                                   "Power consumption is {0}", info=state)
            await self.log.error_check(device["hive_id"], "Extra", online)
            final = state
            Data.NODES[device["hive_id"]]["Power"] = final
        else:
            await self.log.error_check(device["hive_id"], "ERROR", "Failed")

        return final if final is None else Data.NODES[device["hive_id"]]["Power"]

    async def turn_on(self, device):
        """Set smart plug to turn on."""
        await self.log.log(device["hive_id"], "Extra", "Powering switch")
        final = False

        if device["hive_id"] in Data.products:
            await self.session.hive_api_logon()
            data = Data.products[device["hive_id"]]
            resp = await self.hive.set_state(Data.sess_id,
                                             data["type"],
                                             data["id"],
                                             status="ON")
            if resp["original"] == 200:
                final = True
                await self.session.get_devices(device["hive_id"])
                await self.log.log(device["hive_id"], "API", "Switched on - API response 200")
            else:
                await self.log.error_check(
                    device["hive_id"], "ERROR", "Failed_API", resp=resp["original"])

        return final

    async def turn_off(self, device):
        """Set smart plug to turn off."""
        await self.log.log(device["hive_id"], "Extra", "Turning off switch.")
        final = False

        if device["hive_id"] in Data.products:
            await self.session.hive_api_logon()
            data = Data.products[device["hive_id"]]
            resp = await self.hive.set_state(
                Data.sess_id,
                data["type"],
                data["id"],
                status="OFF")
            if resp["original"] == 200:
                final = True
                await self.session.get_devices(device["hive_id"])
                await self.log.log(device["hive_id"], "API", "Switch off - API response 200")
            else:
                await self.log.error_check(
                    device["hive_id"], "ERROR", "Failed_API", resp=resp["original"])

        return final



"""Hive Data Module."""
import datetime


class Data:
    """Hive Data"""

    NODE_INTERVAL_DEFAULT = 120
    WEATHER_INTERVAL_DEFAULT = 600

    # API Data
    products = {}
    devices = {}
    actions = {}
    user = {}
    NODES = {"Preheader": {"Header": "HeaderText"}}
    MODE = []
    BATTERY = []
    HIVETOHA = {
        "Attribute": {True: "Online", False: "Offline"},
        "Boost": {None: "OFF", False: "OFF"},
        "Heating": {False: "OFF"},
        "Hotwater": {"MANUAL": "ON", None: "OFF", False: "OFF"},
        "Hub": {
            "Status": {True: 1, False: 0},
            "Smoke": {True: 1, False: 0},
            "Dog": {True: 1, False: 0},
            "Glass": {True: 1, False: 0},
        },
        "Light": {"ON": True, "OFF": False},
        "Sensor": {"OPEN": True, "CLOSED": False, True: "Online", False: "Offline"},
        "Switch": {"ON": True, "OFF": False},
    }
    HIVE_TYPES = {
        "Hub": ["hub", "sense"],
        "Thermo": ["thermostatui", "trv"],
        "Heating": ["heating", "trvcontrol"],
        "Hotwater": ["hotwater"],
        "Plug": ["activeplug"],
        "Light": ["warmwhitelight", "tuneablelight", "colourtuneablelight"],
        "Sensor": ["motionsensor", "contactsensor"],
    }
    sensor_commands = {
        "SMOKE_CO": "self.hub.hub_smoke(device)",
        "DOG_BARK": "self.hub.hub_dog_bark(device)",
        "GLASS_BREAK": "self.hub.hub_glass(device)",
        "CurrentTemperature": "self.heating.current_temperature(device)",
        "TargetTemperature": "self.heating.target_temperature(device)",
        "Heating_State": "self.heating.get_state(device)",
        "Heating_Mode": "self.heating.get_mode(device)",
        "Heating_Boost": "self.heating.boost(device)",
        "Hotwater_State": "self.hotwater.get_state(device)",
        "Hotwater_Mode": "self.hotwater.get_mode(device)",
        "Hotwater_Boost": "self.hotwater.get_boost(device)",
        "Battery": 'self.attributes.battery(device["device_id"])',
        "Mode": 'self.attributes.get_mode(device["device_id"])',
        "Availability": 'self.online(device)',
        "Connectivity": "self.online(device)",
        "Weather": "self.weather.temperature(device)"}

    # Session Data
    sess_id = None
    s_token = False
    s_logon_datetime = datetime.datetime.now()
    s_username = ""
    s_password = ""
    s_interval_seconds = NODE_INTERVAL_DEFAULT
    s_last_update = datetime.datetime(2017, 1, 1, 12, 0, 0)
    s_file = False

    # Weather data
    w_last_update = datetime.datetime(2017, 1, 1, 12, 0, 0)
    w_nodeid = ""
    w_icon = ""
    w_description = ""
    w_interval_seconds = WEATHER_INTERVAL_DEFAULT
    w_temperature_unit = ""
    w_temperature_value = 0.00

    # Platform data
    p_minmax = {}

    # Logging data
    l_o_folder = ""
    l_o_file = ""
    l_files = {
        "All": "log.all",
        "Action": "log.aciton",
        "Attribute": "log.attribute",
        "API": "log.api",
        "API_CORE": "log.api_core",
        "ERROR": "log.error",
        "Extra": "log.extra",
        "Heating": "log.heating",
        "Hotwater": "log.hotwater",
        "Light": "log.light",
        "Sensor": "log.sensor",
        "Session": "log.session",
        "Switch": "log.switch",
    }
    l_values = {}



"""Define a client to interact with the Hive APIs."""
import logging
from typing import Optional

from aiohttp import ClientSession
from .hive_async_api import Hive_Async

_LOGGER = logging.getLogger(__name__)

DEFAULT_API_VERSION = 1


class Client:  # pylint: disable=too-few-public-methods
    """Define the client."""

    def __init__(self, session: Optional[ClientSession] = None,):
        """Initialize."""
        Data.client = Hive_Async(session)



""""Hive Heating Module. """
import asyncio
from datetime import datetime
from typing import Optional
from aiohttp import ClientSession
from .hive_session import Session
from .hive_data import Data
from .custom_logging import Logger
from .device_attributes import Attributes
from .hive_async_api import Hive_Async


class Heating:
    """Hive Heating Code."""

    def __init__(self, websession: Optional[ClientSession] = None):
        """Initialise."""
        self.hive = Hive_Async(websession)
        self.session = Session(websession)
        self.log = Logger()
        self.attr = Attributes()
        self.type = "Heating"

    async def get_heating(self, device):
        await self.log.log(device["hive_id"], self.type, "Getting heating data.")
        online = await self.attr.online_offline(device["device_id"])
        error = await self.log.error_check(device["hive_id"], self.type, online)

        dev_data = {}
        if device["device_id"] in Data.devices:
            data = Data.devices[device["device_id"]]
            dev_data = {"hive_id": device["hive_id"],
                        "hive_name": device["hive_name"],
                        "hive_type": device["hive_type"],
                        "ha_name": device["ha_name"],
                        "ha_type": device["ha_type"],
                        "device_id": device["device_id"],
                        "device_name": device["device_name"],
                        "temperatureunit": device["temperatureunit"],
                        "min_temp": await self.min_temperature(device),
                        "max_temp": await self.max_temperature(device),
                        "current_temperature": await self.current_temperature(device),
                        "target_temperature": await self.target_temperature(device),
                        "action": await self.current_operation(device),
                        "mode": await self.get_mode(device),
                        "boost": await self.boost(device),
                        "device_data": data.get("props", None),
                        "parent_device": data.get("parent", None),
                        "custom": device.get("custom", None),
                        "attributes": await self.attr.state_attributes(device["device_id"],
                                                                       device["hive_type"])
                        }
        if not error:
            await self.log.log(device["hive_id"], self.type,
                               "Device update {0}", info=dev_data)

        return dev_data

    @staticmethod
    async def min_temperature(device):
        """Get heating minimum target temperature."""
        if device["hive_type"] == "nathermostat":
            return Data.products[device["hive_id"]]["props"]["minHeat"]
        return 5

    @staticmethod
    async def max_temperature(device):
        """Get heating maximum target temperature."""
        if device["hive_type"] == "nathermostat":
            return Data.products[device["hive_id"]]["props"]["maxHeat"]
        return 32

    async def current_temperature(self, device):
        """Get heating current temperature."""
        from datetime import datetime

        await self.log.log(device["hive_id"], "Extra", "Getting current temp")
        online = await self.attr.online_offline(device["device_id"])
        f_state = None
        state = None
        final = None

        if device["hive_id"] in Data.products:
            if online:
                data = Data.products[device["hive_id"]]
                state = data["props"]["temperature"]

                if device["hive_id"] in Data.p_minmax:
                    if Data.p_minmax[device["hive_id"]]["TodayDate"] != datetime.date(
                        datetime.now()
                    ):
                        Data.p_minmax[device["hive_id"]]["TodayMin"] = 1000
                        Data.p_minmax[device["hive_id"]]["TodayMax"] = -1000
                        Data.p_minmax[device["hive_id"]]["TodayDate"] = datetime.date(
                            datetime.now())

                        if state < Data.p_minmax[device["hive_id"]]["TodayMin"]:
                            Data.p_minmax[device["hive_id"]
                                          ]["TodayMin"] = state

                        if state > Data.p_minmax[device["hive_id"]]["TodayMax"]:
                            Data.p_minmax[device["hive_id"]
                                          ]["TodayMax"] = state

                        if state < Data.p_minmax[device["hive_id"]]["RestartMin"]:
                            Data.p_minmax[device["hive_id"]
                                          ]["RestartMin"] = state

                        if state > Data.p_minmax[device["hive_id"]]["RestartMax"]:
                            Data.p_minmax[device["hive_id"]
                                          ]["RestartMax"] = state
                else:
                    data = {
                        "TodayMin": state,
                        "TodayMax": state,
                        "TodayDate": str(datetime.date(datetime.now())),
                        "RestartMin": state,
                        "RestartMax": state,
                    }
                    Data.p_minmax[device["hive_id"]] = data
                f_state = round(float(state), 1)
                await self.log.log(device["hive_id"], "Extra",
                                   "Current Temp is {0}", info=str(state))
            await self.log.error_check(device["hive_id"], "Extra", online)
            final = f_state
            Data.NODES[device["hive_id"]]["CurrentTemp"] = final
        else:
            await self.log.error_check(device["hive_id"], "ERROR", "Failed")

        return final if final is None else Data.NODES[device["hive_id"]]["CurrentTemp"]

    async def minmax_temperatures(self, device):
        """Min/Max Temp"""
        await self.log.log(device["hive_id"], "Extra", "Getting Min/Max temp")
        online = await self.attr.online_offline(device["hive_id"])
        state = None
        final = None

        if device["hive_id"] in Data.p_minmax:
            if online:
                state = Data.p_minmax[device["hive_id"]]
                await self.log.log(device["hive_id"], "Extra", "Min/Max is {0}", info=state)
            await self.log.error_check(device["hive_id"], "Extra", online)
            final = state
            Data.NODES[device["hive_id"]]["minmax"] = final
        else:
            await self.log.error_check(device["hive_id"], "ERROR", "Failed")

        return final if final is None else Data.NODES[device["hive_id"]]["minmax"]

    async def target_temperature(self, device):
        """Get heating target temperature."""
        await self.log.log(device["hive_id"], "Extra", "Getting target temp")
        online = await self.attr.online_offline(device["device_id"])
        state = None
        final = None

        if device["hive_id"] in Data.products:
            if online:
                data = Data.products[device["hive_id"]]
                state = round(float(data["state"].get("target", None)))
                state = round(float(data["state"].get("heat", state)))
                await self.log.log(device["hive_id"], "Extra",
                                   "Target temp is {0}", info=str(state))
            await self.log.error_check(device["hive_id"], "Extra", online)
            final = state
            Data.NODES[device["hive_id"]]["TargetTemp"] = final
        else:
            await self.log.error_check(device["hive_id"], "ERROR", "Failed")

        return final if final is None else Data.NODES[device["hive_id"]]["TargetTemp"]

    async def get_mode(self, device):
        """Get heating current mode."""
        await self.log.log(device["hive_id"], "Extra", "Getting mode")
        online = await self.attr.online_offline(device["device_id"])
        state = None
        final = None

        if device["hive_id"] in Data.products:
            if online:
                data = Data.products[device["hive_id"]]
                state = data["state"]["mode"]
                if state == "BOOST":
                    state = data["props"]["previous"]["mode"]
                await self.log.log(device["hive_id"], "Extra", "Mode is {0}", info=str(state))
            await self.log.error_check(device["hive_id"], "Extra", online)
            final = Data.HIVETOHA[self.type].get(state, state)
            Data.NODES[device["hive_id"]]["Mode"] = final
        else:
            await self.log.error_check(device["hive_id"], "ERROR", "Failed")

        return final if final is None else Data.NODES[device["hive_id"]]["Mode"]

    async def get_state(self, device):
        """Get heating current state."""
        await self.log.log(device["hive_id"], "Extra", "Getting state")
        online = await self.attr.online_offline(device["device_id"])
        state = None
        final = None

        if device["hive_id"] in Data.products:
            if online:
                current_temp = await self.current_temperature(device)
                target_temp = await self.target_temperature(device)
                if current_temp < target_temp:
                    state = "ON"
                else:
                    state = "OFF"
                await self.log.log(device["hive_id"], "Extra", "State is {0}", info=str(state))
            await self.log.error_check(device["hive_id"], "Extra", online)
            final = Data.HIVETOHA[self.type].get(state, state)
            Data.NODES[device["hive_id"]]["State"] = final
        else:
            await self.log.error_check(device["hive_id"], "ERROR", "Failed")

        return final if final is None else Data.NODES[device["hive_id"]]["State"]

    async def current_operation(self, device):
        """Get heating current operation."""
        await self.log.log(device["hive_id"], "Extra", "Getting current operation")
        online = await self.attr.online_offline(device["device_id"])
        state = None
        final = None

        if device["hive_id"] in Data.products:
            if online:
                data = Data.products[device["hive_id"]]
                state = data["props"]["working"]
                await self.log.log(device["hive_id"], "Extra",
                                   "Current operation is {0}", info=str(state))
            await self.log.error_check(device["hive_id"], "Extra", online)
            final = state
            Data.NODES[device["hive_id"]]["Current_Operation"] = final
        else:
            await self.log.error_check(device["hive_id"], "ERROR", "Failed")

        return final if final is None else Data.NODES[device["hive_id"]]["Current_Operation"]

    async def boost(self, device):
        """Get heating boost current status."""
        await self.log.log(device["hive_id"], "Extra", "Getting boost status")
        online = await self.attr.online_offline(device["device_id"])
        state = None
        final = None

        if device["hive_id"] in Data.products:
            if online:
                data = Data.products[device["hive_id"]]
                state = Data.HIVETOHA["Boost"].get(
                    data["state"].get("boost", False), "ON")
                await self.log.log(device["hive_id"], "Extra",
                                   "Boost state is {0}", info=str(state))
            await self.log.error_check(device["hive_id"], "Extra", online)
            final = state
            Data.NODES[device["hive_id"]]["Boost"] = final
        else:
            await self.log.error_check(device["hive_id"], "ERROR", "Failed")

        return final if final is None else Data.NODES[device["hive_id"]]["Boost"]

    async def get_boost_time(self, device):
        """Get heating boost time remaining."""
        if await self.boost(device) == "ON":
            await self.log.log(device["hive_id"], "Extra", "Getting boost time")
            online = await self.attr.online_offline(device["device_id"])
            state = None
            final = None

            if device["hive_id"] in Data.products:
                if online:
                    data = Data.products[device["hive_id"]]
                    state = data["state"]["boost"]
                    await self.log.log(
                        device["hive_id"], self.type, "Time left on boost is {0}", info=str(state)
                    )
                await self.log.error_check(device["hive_id"], "Extra", online)
                final = state
                Data.NODES[device["hive_id"]]["Boost_Time"] = final
            else:
                await self.log.error_check(device["hive_id"], "ERROR", "Failed")

            return final if final is None else Data.NODES[device["hive_id"]]["Boost_Time"]
        return None

    @staticmethod
    async def get_operation_modes():
        """Get heating list of possible modes."""
        return ["SCHEDULE", "MANUAL", "OFF"]

    async def get_schedule_now_next_later(self, device):
        """Hive get heating schedule now, next and later."""
        await self.log.log(device["hive_id"], "Extra", "Getting schedule")
        state = await self.attr.online_offline(device["device_id"])
        current_mode = await self.get_mode(device)
        state = None
        final = None

        if device["hive_id"] in Data.products:
            if state != "Offline" and current_mode == "SCHEDULE":
                data = Data.products[device["hive_id"]]
                state = await self.session.p_get_schedule_nnl(
                    Session(), (data["state"]["schedule"])
                )
                await self.log.log(device["hive_id"], "Extra",
                                   "Schedule is {0}", info=str(state))
            await self.log.error_check(device["hive_id"], "Extra", state)
            final = state
            Data.NODES[device["hive_id"]]["snnl"] = final
        else:
            await self.log.error_check(device["hive_id"], "ERROR", "Failed")

        return final if final is None else Data.NODES[device["hive_id"]]["snnl"]

    async def set_target_temperature(self, device, new_temp):
        """Set heating target temperature."""
        final = False

        if device["hive_id"] in Data.products:
            await self.session.hive_api_logon()
            data = Data.products[device["hive_id"]]
            resp = await self.hive.set_state(Data.sess_id, data["type"],
                                             device["hive_id"],  target=new_temp)

            if resp["original"] == 200:
                await self.session.get_devices(device["hive_id"])
                final = True
                await self.log.log(device["hive_id"], "API", "Temperature set - API response 200")
            else:
                await self.log.error_check(
                    device["hive_id"], "ERROR", "Failed_API", resp=resp["original"])

        return final

    async def set_mode(self, device, new_mode):
        """Set heating mode."""
        await self.session.hive_api_logon()
        final = False

        if device["hive_id"] in Data.products:
            data = Data.products[device["hive_id"]]
            resp = await self.hive.set_state(Data.sess_id, data["type"],
                                             device["hive_id"], mode=new_mode)

            if resp["original"] == 200:
                await self.session.get_devices(device["hive_id"])
                final = True
                await self.log.log(device["hive_id"], "API", "Mode updated - API response 200")
            else:
                await self.log.error_check(
                    device["hive_id"], "ERROR", "Failed_API", resp=resp["original"])

        return final

    async def turn_boost_on(self, device, mins, temp):
        """Turn heating boost on."""
        if mins > 0 and temp >= await self.min_temperature(device):
            if temp <= await self.max_temperature(device):
                await self.log.log(device["hive_id"], "Extra", "Enabling boost for {0}")
                await self.session.hive_api_logon()
                final = False

                if device["hive_id"] in Data.products:
                    data = Data.products[device["hive_id"]]
                    resp = await self.hive.set_state(Data.sess_id, data["type"],
                                                     device["hive_id"], mode="BOOST", boost=mins, target=temp)

                    if resp["original"] == 200:
                        await self.session.get_devices(device["hive_id"])
                        final = True
                        await self.log.log(
                            device["hive_id"], "API", "Boost enabled - " +
                            "API response 200"
                        )
                    else:
                        await self.log.error_check(
                            device["hive_id"], "ERROR", "Failed_API", resp=resp["original"]
                        )

                return final
        return None

    async def turn_boost_off(self, device):
        """Turn heating boost off."""
        await self.log.log(self.type, "Disabling boost for {0}", device["hive_id"])
        final = False

        if device["hive_id"] in Data.products:
            await self.session.hive_api_logon()
            data = Data.products[device["hive_id"]]
            await self.session.get_devices(device["hive_id"])
            if await self.boost(device) == "ON":
                prev_mode = data["props"]["previous"]["mode"]
                if prev_mode == "MANUAL" or prev_mode == "OFF":
                    pre_temp = data["props"]["previous"].get("target", 7)
                    resp = await self.hive.set_state(Data.sess_id, data["type"],
                                                     device["hive_id"],
                                                     mode=prev_mode,
                                                     target=pre_temp)
                else:
                    resp = await self.hive.set_state(Data.sess_id, data["type"],
                                                     device["hive_id"], mode=prev_mode)
                if resp["original"] == 200:
                    await self.session.get_devices(device["hive_id"])
                    final = True
                    await self.log.log(
                        device["hive_id"], "API", "Boost disabled - " + "API response 200")
                else:
                    await self.log.error_check(
                        device["hive_id"], "ERROR", "Failed_API", resp=resp["original"]
                    )

        return final



"""__init__.py"""
from .hive_session import Session
from .action import Action
from .client import Client
from .custom_logging import Logger
from .hive_api import Hive
from .device_attributes import Attributes
from .heating import Heating
from .hive_async_api import Hive_Async
from .hive_data import Data
from .hotwater import Hotwater
from .hub import Hub
from .light import Light
from .plug import Plug
from .sensor import Sensor
from .weather import Weather



"""Custom Logging Module."""
import os
from datetime import datetime

from .hive_data import Data


class Logger:
    """Custom Logging Code."""

    def __init__(self):
        """Logger Initialization"""

    @staticmethod
    async def check_logging():
        """Check Logging Active"""
        Data.l_o_folder = os.path.expanduser("~") + "/pyhiveapi"
        Data.l_o_file = Data.l_o_folder + "/pyhiveapi.log"
        count = 0
        try:

            statinfo = os.stat(os.path.isfile(Data.l_o_file))
            if os.path.isfile(Data.l_o_file) and statinfo.st_size > 1048576:
                os.remove(Data.l_o_file)

            if os.path.isdir(Data.l_o_folder):
                for name in Data.l_files:
                    loc = Data.l_o_folder + "/" + Data.l_files[name]
                    if os.path.isfile(loc):
                        Data.l_values.update({name: True})
                        Data.l_values.update({"enabled": True})
                    elif os.path.isfile(loc) is False:
                        count += 1
                        if count == len(Data.l_files):
                            Data.l_values = {}
        except FileNotFoundError:
            Data.l_values = {}

    @staticmethod
    async def log(n_id, l_type, new_message, **kwargs):
        """Output new log entry if logging is turned on."""
        product_data = Data.products.get(n_id, {})
        name = product_data.get("state", {}).get("name", n_id)
        data = kwargs.get("info", None)
        values = Data.l_values
        if n_id == "No_ID":
            name = "Hive"
        final = False
        if "_" in l_type:
            nxt = l_type.split("_")
            for i in nxt:
                if i in Data.l_values or "All" in Data.l_values and l_type != "Extra":
                    if Data.l_values["enabled"]:
                        final = True
                        break
        elif l_type in values or "All" in values and l_type != "Extra":
            if Data.l_values["enabled"]:
                final = True
        elif l_type in values and Data.l_values["enabled"]:
            final = True

        if final and new_message is not None:
            try:
                l_file = open(Data.l_o_file, "a")
                l_file.write(
                    datetime.now().strftime("%d-%b-%Y %H:%M:%S")
                    + " - "
                    + l_type
                    + " - "
                    + name
                    + " : "
                    + new_message.format(data)
                    + "\n"
                )
                l_file.close()
            except FileNotFoundError:
                pass
        else:
            pass

    async def error_check(self, n_id, n_type, error_type, **kwargs):
        """Error has occurred."""
        message = None
        new_data = None
        result = False
        if error_type == False:
            message = "Device offline could not update entity."
            result = True
        elif error_type == "Failed":
            message = "ERROR - No data found for device."
            result = True
        elif error_type == "Failed_API":
            new_data = str(kwargs.get("resp"))
            message = "ERROR - Received {0} response from API."
            result = True
        await self.log(n_id, n_type, message, info=new_data)
        return result



""" Hive Session Module."""
import operator
import threading
import os
import inspect
import time
import json
import asyncio
from typing import Optional
from aiohttp import ClientSession
from datetime import datetime, timedelta


from .hive_data import Data
from .custom_logging import Logger
from .hive_async_api import Hive_Async
from .weather import Weather
from .hub import Hub
from .device_attributes import Attributes

MINUTES_BETWEEN_LOGONS = 15


class Session:
    """Hive Session Code"""

    def __init__(self, websession: Optional[ClientSession] = None):
        """Initialise the base variable values."""
        self.update_lock = asyncio.Lock()
        self.api_lock = asyncio.Lock()
        self.hive = Hive_Async(websession)
        self.log = Logger()
        self.attr = Attributes()
        self.devices = {}
        self.type = "Session"

    async def open_file(self, file):
        path = os.path.expanduser("~") + "/pyhiveapi" + '/' + file
        with open(path, 'r') as j:
            data = json.loads(j.read())

        return data

    async def add_list(self, name, data, **kwargs):
        """Add entity to the list"""
        formatted_data = {}
        try:
            formatted_data = {"hive_id": data.get("id", ""),
                              "hive_name": data.get("state", {}).get("name", ""),
                              "hive_type": data.get("type", ""),
                              "ha_type": name
                              }
            if kwargs.get("ha_name", "FALSE")[0] == " ":
                kwargs["ha_name"] = data.get("state", {}).get(
                    "name", "") + kwargs["ha_name"]
            else:
                formatted_data["ha_name"] = formatted_data["hive_name"]
            formatted_data.update(kwargs)
        except (KeyError):
            await self.log.log("No_ID", self.type, "Could not setup device - " + str(data))

        self.devices[name].append(formatted_data)

    async def update_interval(self, new_interval):
        "Update the scan interval."
        if new_interval < 15:
            new_interval = 15
        Data.s_interval_seconds = new_interval
        text = "Scan interval updated to " + str(Data.s_interval_seconds)
        await self.log.log("No_ID", self.type, text)

    async def hive_api_logon(self):
        """Log in to the Hive API and get the Session Data."""

        if True in (Data.s_file, Data.s_token):
            return None
        else:
            await self.log.log("No_ID", self.type, "Checking if login is required.")
            c_time = datetime.now()
            l_logon_secs = (c_time - Data.s_logon_datetime).total_seconds()
            l_logon_mins = int(round(l_logon_secs / 60))

            if l_logon_mins >= MINUTES_BETWEEN_LOGONS or Data.sess_id is None:
                await self.log.log("No_ID", self.type, "Attempting to login to Hive.")
                login_details_found = False
                try_finished = False

                resp_p = await self.hive.login(Data.s_username, Data.s_password)

                if resp_p["original"] == 200:
                    info = resp_p["parsed"]
                    if "token" in info and "user" in info and "platform" in info:
                        Data.sess_id = info["token"]
                        Data.s_logon_datetime = datetime.now()
                        login_details_found = True

                        self.hive.urls.update(
                            {"base": info["platform"]["endpoint"]})

    async def update_data(self, device):
        """Get latest data for Hive nodes - rate limiting."""
        await self.log.check_logging()
        await self.update_lock.acquire()
        try:
            updated = False
            ct = datetime.now()
            last_update_secs = (ct - Data.s_last_update).total_seconds()
            if last_update_secs >= Data.s_interval_seconds:
                updated = await self.get_devices(device["hive_id"])

            w_last_update_secs = (ct - Data.w_last_update).total_seconds()
            if w_last_update_secs >= Data.w_interval_seconds:
                updated = await self.get_weather()
        finally:
            self.update_lock.release()

        return updated

    async def get_devices(self, n_id):
        """Get latest data for Hive nodes."""
        get_nodes_successful = False
        api_resp_d = None

        if self.api_lock.locked():
            return True

        try:
            await self.api_lock.acquire()
            await asyncio.sleep(1)
            if Data.s_file:
                await self.log.log(n_id, "API", "Using file")
                api_resp_d = await self.open_file("data.json")
            elif Data.sess_id is not None:
                await self.log.log(n_id, "Session", "Getting hive device info")
                await self.hive_api_logon()
                api_resp_d = await self.hive.get_all(Data.sess_id)
                if api_resp_d["original"] == 200:
                    await self.log.log(n_id, "API", "API response 200")
            else:
                await self.log.error_check(
                    n_id, "ERROR", "Failed_API", resp=api_resp_d["original"])
                asyncio.CancelledError('Could not get response from the API')

            api_resp_p = api_resp_d["parsed"]

            for hive_type in api_resp_p:
                if hive_type == "user":
                    Data.user = api_resp_p[hive_type]
                if hive_type == "products":
                    Data.products = {}
                    for a_product in api_resp_p[hive_type]:
                        Data.products.update({a_product["id"]: a_product})
                        Data.NODES.update({a_product["id"]: {"": ""}})
                if hive_type == "devices":
                    Data.devices = {}
                    for a_device in api_resp_p[hive_type]:
                        Data.devices.update({a_device["id"]: a_device})
                        Data.NODES.update({a_device["id"]: {"": ""}})
                if hive_type == "actions":
                    Data.actions = {}
                    for a_action in api_resp_p[hive_type]:
                        Data.actions.update({a_action["id"]: a_action})
                        Data.NODES.update({a_action["id"]: {"": ""}})

            Data.s_last_update = datetime.now()
            get_nodes_successful = True
        except (IOError, RuntimeError, ZeroDivisionError, ConnectionError):
            await self.log.log("No_ID", "Core_API", "Api didn't receive any data")
            get_nodes_successful = False
        finally:
            self.api_lock.release()

        return get_nodes_successful

    async def get_weather(self):
        """Get latest weather data from Hive."""
        get_weather_successful = False
        current_time = datetime.now()

        try:
            if Data.sess_id is not None and Data.s_file is not True:
                await self.log.log("No_ID", self.type, "Getting Hive weather info.")
                await self.hive_api_logon()
                weather_url = (
                    "?postcode=" +
                    Data.user.get("postcode", '') +
                    "&country=" + Data.user.get("countryCode", '')
                )

                resp = await self.hive.get_weather(weather_url)
                if resp["original"] == 200:
                    await self.log.log("No_ID", "API", "API response 200")
                    Data.w_icon = resp["parsed"]["weather"]["icon"]
                    Data.w_description = resp["parsed"]["weather"]["description"]
                    Data.w_temperature_unit = resp["parsed"]["weather"]["temperature"]["unit"]
                    Data.w_temperature_value = resp["parsed"]["weather"]["temperature"]["value"]
                    Data.w_nodeid = "HiveWeather"
                    Data.w_last_update = current_time
                    get_weather_successful = True

        except (IOError, RuntimeError, ZeroDivisionError):
            get_weather_successful = False

        return get_weather_successful

    @staticmethod
    async def p_minutes_to_time(minutes_to_convert):
        """Convert minutes string to datetime."""
        hours_converted, minutes_converted = divmod(minutes_to_convert, 60)
        converted_time = datetime.strptime(
            str(hours_converted) + ":" + str(minutes_converted), "%H:%M"
        )
        converted_time_string = converted_time.strftime("%H:%M")
        return converted_time_string

    @staticmethod
    async def p_get_schedule_nnl(self, hive_api_schedule):
        """Get the schedule now, next and later of a given nodes schedule."""
        schedule_now_and_next = {}
        date_time_now = datetime.now()
        date_time_now_day_int = date_time_now.today().weekday()

        days_t = (
            "monday",
            "tuesday",
            "wednesday",
            "thursday",
            "friday",
            "saturday",
            "sunday",
        )

        days_rolling_list = list(days_t[date_time_now_day_int:] + days_t)[:7]

        full_schedule_list = []

        for day_index in range(0, len(days_rolling_list)):
            current_day_schedule = hive_api_schedule[days_rolling_list[day_index]]
            current_day_schedule_sorted = sorted(
                current_day_schedule, key=operator.itemgetter("start"), reverse=False
            )

            for current_slot in range(0, len(current_day_schedule_sorted)):
                current_slot_custom = current_day_schedule_sorted[current_slot]

                slot_date = datetime.now() + timedelta(days=day_index)
                slot_time = await self.p_minutes_to_time(
                    current_slot_custom["start"])
                slot_time_date_s = slot_date.strftime(
                    "%d-%m-%Y") + " " + slot_time
                slot_time_date_dt = datetime.strptime(
                    slot_time_date_s, "%d-%m-%Y %H:%M"
                )
                if slot_time_date_dt <= date_time_now:
                    slot_time_date_dt = slot_time_date_dt + timedelta(days=7)

                current_slot_custom["Start_DateTime"] = slot_time_date_dt
                full_schedule_list.append(current_slot_custom)

        fsl_sorted = sorted(
            full_schedule_list, key=operator.itemgetter("Start_DateTime"), reverse=False
        )

        schedule_now = fsl_sorted[-1]
        schedule_next = fsl_sorted[0]
        schedule_later = fsl_sorted[1]

        schedule_now["Start_DateTime"] = schedule_now["Start_DateTime"] - timedelta(
            days=7
        )

        schedule_now["End_DateTime"] = schedule_next["Start_DateTime"]
        schedule_next["End_DateTime"] = schedule_later["Start_DateTime"]
        schedule_later["End_DateTime"] = fsl_sorted[2]["Start_DateTime"]

        schedule_now_and_next["now"] = schedule_now
        schedule_now_and_next["next"] = schedule_next
        schedule_now_and_next["later"] = schedule_later

        return schedule_now_and_next

    async def start_session(self, interval, token, username, password):
        """Setup the Hive platform."""
        await self.log.check_logging()
        await self.log.log("No_ID", self.type, "Initialising Hive Component.")

        loc = None
        loc = os.path.expanduser("~") + "/pyhiveapi/use_file"
        Token = token
        Username = username
        Password = password
        await self.update_interval(interval)
        if os.path.isfile(loc):
            Data.s_file = True
            Data.sess_id = "File"
        elif Username is not None and Password is not None and Token is None:
            await self.log.log("No_ID", self.type,
                               "Using credentials for autentication.")
            Data.s_username = Username
            Data.s_password = Password
            await self.hive_api_logon()
        elif Token is not None:
            await self.log.log("No_ID", self.type,
                               "Using a long lived token for authentiction.")
            Data.sess_id = Token
            Data.s_token = True
        else:
            raise ConnectionError(
                "Cannot connect to Hive without token or credentials")

        if Data.sess_id is not None:
            await self.get_devices("No_ID")
            await self.get_weather()

        if Data.devices is None or Data.products is None:
            await self.log.log("No_ID", self.type, "Failed to get data")
            return None

        await self.log.log("No_ID", self.type, "Creating list of devices")
        self.devices["binary_sensor"] = []
        self.devices["climate"] = []
        self.devices["light"] = []
        self.devices["sensor"] = []
        self.devices["switch"] = []
        self.devices["water_heater"] = []

        for a_product in Data.products:
            p = Data.products[a_product]
            if Data.products[a_product]["type"] == "sense":
                await self.add_list("binary_sensor", p, ha_name="Glass Detection",
                                    hive_type="GLASS_BREAK", hub_id=p["parent"],
                                    device_id=p["parent"], device_name=p["state"]["name"])
                await self.add_list("binary_sensor", p, ha_name="Smoke Detection",
                                    hive_type="SMOKE_CO", hub_id=p["parent"],
                                    device_id=p["parent"], device_name=p["state"]["name"])
                await self.add_list("binary_sensor", p, ha_name="Dog Bark Detection",
                                    hive_type="DOG_BARK", hub_id=p["parent"],
                                    device_id=p["parent"], device_name=p["state"]["name"])
            if Data.products[a_product]["type"] in Data.HIVE_TYPES["Heating"]:
                device_id = p["props"].get("zone", p["id"])
                device_name = p["state"].get("name", "Heating")
                for device in Data.devices:
                    try:
                        if Data.devices[device]["type"] in Data.HIVE_TYPES["Thermo"]:
                            device = Data.devices[device]
                            if p["parent"] == device["props"]["zone"]:
                                device_id = device["id"]
                                device_name = device["state"]["name"]
                    except:
                        pass
                Data.MODE.append(p["id"])
                await self.add_list("climate", p, device_id=device_id, device_name=device_name,
                                    temperatureunit=Data.user["temperatureUnit"])
                await self.add_list("sensor", p, ha_name=" Current Temperature",
                                    hive_type="CurrentTemperature", device_id=device_id, device_name=device_name)
                await self.add_list("sensor", p, ha_name=" Target Temperature",
                                    hive_type="TargetTemperature", device_id=device_id, device_name=device_name)
                await self.add_list("sensor", p, ha_name=" State",
                                    hive_type="Heating_State", device_id=device_id, device_name=device_name)
                await self.add_list("sensor", p, ha_name=" Mode",
                                    hive_type="Heating_Mode", device_id=device_id, device_name=device_name)
                await self.add_list("sensor", p, ha_name=" Boost",
                                    hive_type="Heating_Boost", device_id=device_id, device_name=device_name)
            if Data.products[a_product]["type"] in Data.HIVE_TYPES["Hotwater"]:
                device_id = p["props"].get("zone", p["id"])
                device_name = p["state"].get("name", "Hotwater")
                for device in Data.devices:
                    try:
                        if Data.devices[device]["type"] in Data.HIVE_TYPES["Thermo"]:
                            device = Data.devices[device]
                            if p["parent"] == device["props"]["zone"]:
                                device_id = device["id"]
                                device_name = device["state"]["name"]
                    except:
                        pass
                await self.add_list("water_heater", p, device_id=device_id, device_name=device_name)
                await self.add_list("sensor", p, ha_name=" State",
                                    hive_type="Hotwater_State", device_id=device_id, device_name=device_name)
                await self.add_list("sensor", p, ha_name=" Mode",
                                    hive_type="Hotwater_Mode", device_id=device_id, device_name=device_name)
                await self.add_list("sensor", p, ha_name=" Boost",
                                    hive_type="Hotwater_Boost", device_id=device_id, device_name=device_name)
            if Data.products[a_product]["type"] in Data.HIVE_TYPES["Plug"]:
                Data.MODE.append(p["id"])
                await self.add_list("switch", p, device_id=p["id"], device_name=p["state"]["name"])
                await self.add_list("sensor", p, ha_name=" Mode",
                                    hive_type="Mode", device_id=p["id"], device_name=p["state"]["name"])
                await self.add_list("sensor", p, ha_name=" Availability",
                                    hive_type="Availability", device_id=p["id"], device_name=p["state"]["name"])
            if Data.products[a_product]["type"] in Data.HIVE_TYPES["Light"]:
                Data.MODE.append(p["id"])
                await self.add_list("light", p, device_id=p["id"], device_name=p["state"]["name"])
                await self.add_list("sensor", p, ha_name=" Mode",
                                    hive_type="Mode", device_id=p["id"], device_name=p["state"]["name"])
                await self.add_list("sensor", p, ha_name=" Availability",
                                    hive_type="Availability", device_id=p["id"], device_name=p["state"]["name"])
            if Data.products[a_product]["type"] in Data.HIVE_TYPES["Sensor"]:
                await self.add_list("binary_sensor", p, device_id=p["id"], device_name=p["state"]["name"])

        for a_device in Data.devices:
            d = Data.devices[a_device]
            if (
                    Data.devices[a_device]["type"] in Data.HIVE_TYPES["Thermo"]
                    or Data.devices[a_device]["type"] in Data.HIVE_TYPES["Sensor"]):
                Data.BATTERY.append(d["id"])
                await self.add_list("sensor", d, ha_name=" Battery Level",
                                    hive_type="Battery", device_id=d["id"], device_name=d["state"]["name"])
                await self.add_list("sensor", d, ha_name=" Availability",
                                    hive_type="Availability", device_id=d["id"], device_name=d["state"]["name"])
            if Data.devices[a_device]["type"] in Data.HIVE_TYPES["Hub"]:
                await self.add_list("binary_sensor", d, ha_name=" Online Status",
                                    hive_type="Connectivity", device_id=d["id"], device_name=d["state"]["name"])

        for action in Data.actions:
            a = Data.actions[action]
            await self.add_list("switch", a, hive_name=a["name"],
                                ha_name=a["name"],
                                hive_type="action")

        if Data.w_nodeid == "HiveWeather":
            await self.add_list("sensor", {}, hive_id=Data.w_nodeid,
                                hive_name="Hive Weather",
                                ha_name="Hive Weather",
                                hive_type="Weather")

        await self.log.log("No_ID", self.type, "Hive component has initialised")

        return self.devices

    @staticmethod
    async def epochtime(date_time, pattern, action):
        """ date/time conversion to epoch"""
        if action == "to_epoch":
            pattern = "%d.%m.%Y %H:%M:%S"
            epochtime = int(time.mktime(
                time.strptime(str(date_time), pattern)))
            return epochtime
        elif action == "from_epoch":
            date = datetime.fromtimestamp(int(date_time)).strftime(pattern)
            return date



from setuptools import setup, find_packages

setup(
    name="pyhiveapi",
    version="0.3_b3",
    description="A Python library to interface with the Hive API",
    long_description="A Python library to interface with the Hive API",
    url="https://github.com/Rendili/pyhiveapi",
    include_package_data=True,
    author="Rendili",
    author_email="rendili@outlook.com",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        #        'Topic :: Software Development :: API',
        "License :: OSI Approved :: MIT License",
        #        'Programming Language :: Python :: 3.4',
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    python_requires=">=3.5.*",
    keywords="Hive API Library",
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    entry_points={"console_scripts": [
        "pyhiveapi=pyhiveapi.pyhiveapi:Pyhiveapi"]},
    install_requires=["aiohttp", "requests"],
)



""""Hive Hotwater Module. """
import asyncio
from typing import Optional
from aiohttp import ClientSession

from .hive_session import Session
from .hive_data import Data
from .custom_logging import Logger
from .device_attributes import Attributes
from .hive_async_api import Hive_Async


class Hotwater:
    """Hive Hotwater Code."""

    def __init__(self, websession: Optional[ClientSession] = None):
        """Initialise."""
        self.hive = Hive_Async(websession)
        self.session = Session(websession)
        self.log = Logger()
        self.attr = Attributes()
        self.type = "Hotwater"

    async def get_hotwater(self, device):
        """Get light data."""
        await self.log.log(device["hive_id"], self.type, "Getting hot water data.")
        online = await self.attr.online_offline(device["hive_id"])
        error = await self.log.error_check(device["hive_id"], self.type, online)

        dev_data = {}
        if device["device_id"] in Data.devices:
            data = Data.devices[device["device_id"]]
            dev_data = {"hive_id": device["hive_id"],
                        "hive_name": device["hive_name"],
                        "hive_type": device["hive_type"],
                        "ha_name": device["ha_name"],
                        "ha_type": device["ha_type"],
                        "device_id": device["device_id"],
                        "device_name": device["device_name"],
                        "current_operation": await self.get_mode(device),
                        "device_data": data.get("props", None),
                        "parent_device": data.get("parent", None),
                        "custom": device.get("custom", None),
                        "attributes": await self.attr.state_attributes(device["hive_id"],
                                                                       device["hive_type"])
                        }

        if not error:
            await self.log.log(device["hive_id"], self.type,
                               "Device update {0}", info=dev_data)

        return dev_data

    async def get_mode(self, device):
        """Get hotwater current mode."""
        await self.log.log(device["hive_id"], "Extra", "Getting mode")
        online = await self.attr.online_offline(device["hive_id"])
        state = None
        final = None

        if device["hive_id"] in Data.products:
            if online:
                data = Data.products[device["hive_id"]]
                state = data["state"]["mode"]
                if state == "BOOST":
                    state = data["props"]["previous"]["mode"]
                final = Data.HIVETOHA[self.type].get(state, state)
                await self.log.log(device["hive_id"], "Extra", "Mode is {0}", info=final)
            await self.log.error_check(device["hive_id"], "Extra", online)
            final = Data.HIVETOHA[self.type].get(state, state)
            Data.NODES[device["hive_id"]]["Mode"] = final
        else:
            await self.log.error_check(device["hive_id"], "ERROR", "Failed")

        return final if final is None else Data.NODES[device["hive_id"]]["Mode"]

    @staticmethod
    async def get_operation_modes():
        """Get heating list of possible modes."""
        return ["SCHEDULE", "ON", "OFF"]

    async def get_boost(self, device):
        """Get hot water current boost status."""
        await self.log.log(device["hive_id"], "Extra", "Getting boost")
        online = await self.attr.online_offline(device["hive_id"])
        state = None
        final = None

        if device["hive_id"] in Data.products:
            if online:
                data = Data.products[device["hive_id"]]
                state = data["state"]["boost"]
                final = Data.HIVETOHA["Boost"].get(state, "ON")
                await self.log.log(device["hive_id"], "Extra", "Boost is {0}", info=final)
            await self.log.error_check(device["hive_id"], "Extra", online)
            final = Data.HIVETOHA["Boost"].get(state, "ON")
            Data.NODES[device["hive_id"]]["Boost"] = final
        else:
            await self.log.error_check(device["hive_id"], "ERROR", "Failed")

        return final if final is None else Data.NODES[device["hive_id"]]["Boost"]

    async def get_boost_time(self, device):
        """Get hotwater boost time remaining."""
        if await self.get_boost(device["hive_id"]) == "ON":
            await self.log.log(device["hive_id"], "Extra", "Getting boost time")
            online = await self.attr.online_offline(device["hive_id"])
            state = None
        final = None

        if device["hive_id"] in Data.products:
            if online:
                data = Data.products[device["hive_id"]]
                state = data["state"]["boost"]
                await self.log.log(device["hive_id"], "Extra",
                                   "Boost time is {0}", info=state)
            await self.log.error_check(device["hive_id"], "Extra", online)
            final = state
            Data.NODES[device["hive_id"]]["Boost_Time"] = final
        else:
            await self.log.error_check(device["hive_id"], "ERROR", "Failed")

        return final if final is None else Data.NODES[device["hive_id"]]["Boost_Time"]

    async def get_state(self, device):
        """Get hot water current state."""
        await self.log.log(device["hive_id"], "Extra", "Getting state")
        online = await self.attr.online_offline(device["hive_id"])
        state = None
        final = None

        if device["hive_id"] in Data.products:
            if online:
                data = Data.products[device["hive_id"]]
                state = data["state"]["status"]
                mode_current = await self.get_mode(device["hive_id"])
                if mode_current == "SCHEDULE":
                    if await self.get_boost(device["hive_id"]) == "ON":
                        state = "ON"
                    else:
                        snan = self.session.p_get_schedule_nnl(
                            data["state"]["schedule"]
                        )
                        state = snan["now"]["value"]["status"]
            await self.log.error_check(device["hive_id"], "Extra", online)
            final = Data.HIVETOHA[self.type].get(state, state)
            Data.NODES[device["hive_id"]]["State"] = final
        else:
            await self.log.error_check(device["hive_id"], "ERROR", "Failed")

        return final if final is None else Data.NODES[device["hive_id"]]["State"]

    async def get_schedule_now_next_later(self, device):
        """Hive get hotwater schedule now, next and later."""
        await self.log.log(device["hive_id"], "Extra", "Getting schedule info.")
        online = await self.attr.online_offline(device["hive_id"])
        state = None
        final = None

        if device["hive_id"] in Data.products:
            await self.session.hive_api_logon()
            mode_current = await self.get_mode(device)
            if not online and mode_current == "SCHEDULE":
                data = Data.products[device["hive_id"]]
                state = self.session.p_get_schedule_nnl(
                    data["state"]["schedule"])
                final = state
                Data.NODES[device["hive_id"]]["snnl"] = final
                await self.log.log(device["hive_id"], "Extra", "Schedule is {0}", info=final)
            await self.log.error_check(device["hive_id"], "Extra", online)
        else:
            await self.log.error_check(device["hive_id"], "ERROR", "Failed")

        return final

    async def set_mode(self, device, new_mode):
        """Set hot water mode."""
        await self.log.log(device["hive_id"], "Extra", "Setting Mode")
        final = False

        if device["hive_id"] in Data.products:
            await self.session.hive_api_logon()
            data = Data.products[device["hive_id"]]
            resp = await self.hive.set_state(Data.sess_id, data["type"],
                                             device["hive_id"], mode=new_mode)
            if resp["original"] == 200:
                final = True
                await self.session.get_devices(device["hive_id"])
                await self.log.log(
                    device["hive_id"], "API", "Mode set to {0} - API response 200", info=new_mode
                )
            else:
                await self.log.error_check(
                    device["hive_id"], "ERROR", "Failed_API", resp=resp["original"])

            return final

    async def turn_boost_on(self, device, mins):
        """Turn hot water boost on."""
        await self.log.log(device["hive_id"], "Extra", "Turning on boost")
        final = False

        if mins > 0 and device["hive_id"] in Data.products:
            await self.session.hive_api_logon()
            data = Data.products[device["hive_id"]]
            resp = await self.hive.set_state(Data.sess_id, data["type"],
                                             device["hive_id"], mode="BOOST",
                                             boost=mins)
            if resp["original"] == 200:
                final = True
                await self.session.get_devices(device["hive_id"])
                await self.log.log(device["hive_id"], "API", "Boost on - API response 200")
            else:
                await self.log.error_check(
                    device["hive_id"], "ERROR", "Failed_API", resp=resp["original"])

        return final

    async def turn_boost_off(self, device):
        """Turn hot water boost off."""
        await self.log.log(device["hive_id"], "Extra", "Turning off boost")
        final = False

        if device["hive_id"] in Data.products and await self.get_boost(device["hive_id"]) == "ON":
            await self.session.hive_api_logon()
            data = Data.products[device["hive_id"]]
            prev_mode = data["props"]["previous"]["mode"]
            resp = await self.hive.set_state(Data.sess_id, data["type"],
                                             device["hive_id"], mode=prev_mode)
            if resp["original"] == 200:
                await self.session.get_devices(device["hive_id"])
                final = True
                await self.log.log(device["hive_id"], "API", "Boost off - API response 200")
            else:
                await self.log.error_check(
                    device["hive_id"], "ERROR", "Failed_API", resp=resp["original"])

        return final
